/*
    fp: fp_bias.c
    Copyright (C) 2004  Macaulay Institute

    This file is part of fp, a series of tools for looking at floating
    point arithmetic.

    fp is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    fp is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/* fp_bias.c
 *
 * Gary Polhill, 28 April 2004
 *
 * This is a program to test the bias in floating point operations. An
 * operation is repeated a number of times for a given operand over a sequence
 * of othe operands to see if there is any bias in the direction that
 * the floating point result goes.
 *
 * Usage:
 *
 * fp_bias <start number> <sequence> <operator> <operand>
 *
 * <start number> is the initial LH operand
 *
 * <sequence> is the number of floating point numbers from the initial operand
 *   that will be used as other LH operands. If negative then subsequent
 *   LH operands are less than the start operand, otherwise they are greater.
 *
 * <operator> is the operator to use +, -, / or x or *.
 *
 * <operand> is the RH operand, a constant floating point number
 */

#include <ieeefp.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>

typedef enum { PINF = 0, MINF = 1, NEAR = 2, SIZE = 3 } rnd_dir;
static fp_except rnd_dirs[SIZE] = { FP_RP, FP_RM, FP_RN };

int main(int argc, char **argv) {
  double rhop;
  double start;
  double direction;
  char operator;
  long long sequence;
  long long rnd_down = 0, rnd_up = 0;

  if(argc != 5) {
    fprintf(stderr, "Usage: %s <start number> <sequence> <operator> <rh "
	    "operand>\n", argv[0]);
    return 1;
  }

  rhop = atof(argv[4]);
  start = atof(argv[1]);
  sequence = atoll(argv[2]);
  direction = sequence < 0 ? -DBL_MAX : DBL_MAX;
  operator = argv[3][0];
  
  while(sequence != 0) {
    rnd_dir i;
    double results[SIZE];

    for(i = 0; i < SIZE; i++) {
      fpsetround(rnd_dirs[i]);
      switch(operator) {
      case '+':
	results[i] = start + rhop;
	break;
      case '-':
	results[i] = start - rhop;
	break;
      case '/':
	results[i] = start / rhop;
	break;
      case '*':
      case 'x':
	results[i] = start * rhop;
	break;
      default:
	fprintf(stderr, "Invalid setting for operator: %c. Must be one of "
		"+, -, /, or x\n", operator);
	return 2;
      }
    }
    if(results[NEAR] == results[PINF] && results[NEAR] != results[MINF]) {
      rnd_up++;
    }
    if(results[NEAR] == results[MINF] && results[NEAR] != results[PINF]) {
      rnd_down++;
    }
    start = nextafter(start, direction);
    sequence < 0 ? sequence++ : sequence--;
  }

  printf("Up\t%llu\nDown\t%llu\n", rnd_up, rnd_down);

  return 0;
}
