/*
    fp: fp_hacker.c
    Copyright (C) 2003  Macaulay Institute

    This file is part of fp, a series of tools for looking at floating
    point arithmetic.

    fp is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    fp is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/


/*

fp_hacker.c

This is a program to hack floating point numbers. It lets you enter
floating point numbers and find out what their representation is in
binary, and enter binary numbers to find out what floating point
numbers they represent.

*/

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <ieeefp.h>
#include <signal.h>
#include <siginfo.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

#define L2DBL sizeof(double) / sizeof(long)

#define COMPILER "/software/gcc/gcc-3.0.2/bin/gcc"
#define PROGACC 70

static int compiler_exists;
static const double dpinf = 1.0 / 0.0;
static const double dninf = -1.0 / 0.0;
static const float fpinf = 1.0F / 0.0F;
static const float fninf = -1.0F / 0.0F;

typedef union {
  float single_p;
  double double_p;
  long single_b;
  long double_b[L2DBL];
} FP_HACK;

typedef enum { SINGLEFP, DOUBLEFP, SINGLEBIN, DOUBLEBIN } FP_ENTRY;
typedef enum { FPSCANF, FPATOF } FP_READ;
typedef enum { BIN, HEX } BIN_READ;
typedef enum { SET, PLUS, MINUS, MINUSR, TIMES, DIV, DIVR, SQRT, NEXT, PREV,
	       POW, POWR, PLUS2, MINUS2, MINUSR2, TIMES2, DIV2, DIVR2 } FP_OPS;

typedef struct op_list {
  FP_OPS op;
  char *operand;
  struct op_list *next;
} OP_LIST;

OP_LIST *add_op_list(FP_OPS operator, char *num, OP_LIST *head);
void free_op_list(OP_LIST *head);
void print_binary(long number);
void fp_enter(FP_ENTRY entry_type);
void fp_handler(int sig, siginfo_t *sip, void *uap);
void read_binary(FP_HACK *num, FP_ENTRY type, BIN_READ mode);
void read_float(FP_HACK *num, FP_ENTRY type, FP_READ mode, char *entrybuf);
void change_mode(FP_ENTRY *type);
void change_trapping(void);
void change_round(void);
void create_prog(FP_ENTRY type, OP_LIST *commands);

int main(int argc, char **argv) {
  int option = -1;
  struct sigaction fp_action;
  struct stat fstat;

  if(stat(COMPILER, &fstat) == 0) {
    compiler_exists = 1;
  }
  else {
    compiler_exists = 0;
  }

  fpsetmask(
#ifdef FP_X_INV
	    FP_X_INV |
#endif
#ifdef FP_X_DNML
	    FP_X_DNML |
#endif
#ifdef FP_X_DZ
	    FP_X_DZ |
#endif
#ifdef FP_X_OFL
	    FP_X_OFL |
#endif
#ifdef FP_X_UFL
	    FP_X_UFL |
#endif
#ifdef FP_X_IMP
	    FP_X_IMP |
#endif
	    0);
  fp_action.sa_flags = SA_SIGINFO;
  fp_action.sa_sigaction = fp_handler;
  if(sigemptyset(&fp_action.sa_mask) == -1) {
    perror("Setting up floating point error signal mask");
    abort();
  }
  if(sigaction(SIGFPE, &fp_action, NULL) == -1) {
    perror("Setting up floating-point error signal trapping");
    abort();
  }

  while(option == -1) {
    printf("Floating point number hacker -- main menu\n");
    printf("\n\t[1]\tEnter single precision floating point\n");
    printf("\t[2]\tEnter double precision floating point\n");
    printf("\t[3]\tEnter single precision binary number\n");
    printf("\t[4]\tEnter double precision binary number\n");
    printf("\n\t[0]\tExit\n\n> ");
    fflush(stdin);
    scanf("%d", &option);
    switch(option) {
    case 1:
      fp_enter(SINGLEFP);
      break;
    case 2:
      fp_enter(DOUBLEFP);
      break;
    case 3:
      fp_enter(SINGLEBIN);
      break;
    case 4:
      fp_enter(DOUBLEBIN);
      break;
    case 0:
      return 0;
    default:
      printf("Invalid entry: %d\n", option);
    }
    option = -1;
  }
}

OP_LIST *add_op_list(FP_OPS operator, char *num, OP_LIST *head) {
  OP_LIST *new;
  OP_LIST *tail;


  new = (OP_LIST *)malloc(sizeof(OP_LIST));
  if(new == NULL) {
    perror("Memory allocation");
    abort();
  }

  if(head != NULL) {
    tail = head;
    while(tail->next != NULL) {
      tail = tail->next;
    }
    tail->next = new;
  }
  new->op = operator;
  if(num != NULL) {
    new->operand = strdup(num);

    if(new->operand == NULL) {
      perror("Memory allocation");
      abort();
    }
  }
  new->next = NULL;

  return new;
}

void free_op_list(OP_LIST *head) {
  OP_LIST *tail;

  tail = head;
  while(tail != NULL) {
    tail = tail->next;
    if(head->operand != NULL) free(head->operand);
    free(head);
    head = tail;
  }
}

void fp_enter(FP_ENTRY entry_type) {
  FP_HACK fp_num;
  int option = -1;
  int i;
  int sigdig;
  FP_READ fp_mode = FPSCANF;
  BIN_READ bin_mode = BIN;
  OP_LIST *history;
  char fpbuf[1024];

  for(i = 0; i < L2DBL; i++) {
    fp_num.double_b[i] = 0L;
  }

  if(entry_type == SINGLEFP || entry_type == SINGLEBIN) {
    sigdig = FLT_DIG;
  }
  else {
    sigdig = DBL_DIG;
  }

  history = add_op_list(SET, "0.0", NULL);
  while(option == -1) {
    char *binorfp = "binary";
    char *sgordbl = "single";
    int offset;
    FP_HACK temp;
    long overflow;
    fp_except fp_mask = fpgetmask();
    fp_except fp_sticky;

    printf("Number is currently:\n");
    fpsetmask(0);
    switch(entry_type) {
    case SINGLEFP:
      binorfp = "float";
    case SINGLEBIN:
      printf("\tFloat:\t%*g (%.*e)\n\tHex:\t%08lX\n\tBinary:\t", FLT_DIG,
	     (double)fp_num.single_p, sigdig, fp_num.single_p,
	     fp_num.single_b);
      print_binary(fp_num.single_b);
      printf("\n");
      break;
    case DOUBLEFP:
      binorfp = "float";
    case DOUBLEBIN:
      sgordbl = "double";
      printf("\tFloat:\t%*g (%.*le)\n\tHex:\t", DBL_DIG, fp_num.double_p,
	     sigdig, fp_num.double_p);
      for(i = 0; i < L2DBL; i++) {
	printf("%08lX", fp_num.double_b[i]);
      }
      printf("\n\tBinary:\t");
      for(i = 0; i < L2DBL; i++) {
	print_binary(fp_num.double_b[i]);
      }
      printf("\n");
      break;
    }
    printf("\nFloating point number hacker -- entry menu\n\n");
    printf("\t[1]\tEnter new number [%s %s]\n", sgordbl, binorfp);
    if(entry_type == SINGLEFP || entry_type == DOUBLEFP) {
      printf("\t[2]\tAdd entry [%s]\n", sgordbl);
      printf("\t[3]\tSubtract entry [%s]\n", sgordbl);
      printf("\t[4]\tSubtract from entry [%s]\n", sgordbl);
      printf("\t[5]\tMultiply by entry [%s]\n", sgordbl);
      printf("\t[6]\tDivide by entry [%s]\n", sgordbl);
      printf("\t[7]\tDivide entry by [%s]\n", sgordbl);
      printf("\t[8]\tSquare root\n");
      printf("\t[9]\tNext floating point number\n");
      printf("\t[10]\tPrevious floating point number\n");
      printf("\t[11]\tRaise to the power entry [%s]\n", sgordbl);
      printf("\t[12]\tRaise entry to the power [%s]\n", sgordbl);
      printf("\t[13]\tAdd 2^entry [%s]\n", sgordbl);
      printf("\t[14]\tSubtract 2^entry [%s]\n", sgordbl);
      printf("\t[15]\tSubtract from 2^entry [%s]\n", sgordbl);
      printf("\t[16]\tMultiply by 2^entry [%s]\n", sgordbl);
      printf("\t[17]\tDivide by 2^entry [%s]\n", sgordbl);
      printf("\t[18]\tDivide 2^entry by [%s]\n", sgordbl);
      offset = (entry_type == SINGLEFP ? 100 : 200);
    }
    else {
      printf("\t[2]\tAND with number [%s]\n", sgordbl);
      printf("\t[3]\tOR with number [%s]\n", sgordbl);
      printf("\t[4]\tXOR with number [%s]\n", sgordbl);
      printf("\t[5]\tOne's complement\n");
      printf("\t[6]\tLeft shift\n");
      printf("\t[7]\tRight shift\n");
      offset = (entry_type == SINGLEBIN ? 300 : 400);
    }
    printf("\t[50]\tChange IEEE rounding direction\n");
    printf("\t[51]\tChange IEEE floating point exception trapping\n");
    printf("\t[52]\tChange entry mode\n");
    printf("\t[53]\tChange number of significant figures displayed in ()\n");
    if(entry_type == SINGLEFP || entry_type == DOUBLEFP) {
      printf("\t[54]\tToggle the read mode for entry of floating point "
	     "number [%s]\n", (fp_mode == FPSCANF ? "scanf" : "atof"));
      if(compiler_exists) {
	printf("\t[55]\tCreate program to test history\n");
      }
    }
    else {
      printf("\t[54]\tToggle the read mode for entry of binary number [%s]\n",
	     (bin_mode == BIN ? "binary" : "hex"));
    }
    printf("\n\t[0]\tReturn to main menu\n\n> ");
    fflush(stdin);
    scanf("%d", &option);
    if(option < 49 && option > 0) option += offset;
    fpsetsticky(0);
    fpsetmask(fp_mask);
    switch(option) {
    case 101:
    case 201:
      read_float(&fp_num, entry_type, fp_mode, fpbuf);
      free_op_list(history);
      history = add_op_list(SET, fpbuf, NULL);
      break;
    case 301:
    case 401:
      read_binary(&fp_num, entry_type, bin_mode);
      break;
    case 102:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p += temp.single_p;
      add_op_list(PLUS, fpbuf, history);
      break;
    case 202:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p += temp.double_p;
      add_op_list(PLUS, fpbuf, history);
      break;
    case 302:
      read_binary(&temp, entry_type, bin_mode);
      fp_num.single_b &= temp.single_b;
      break;
    case 402:
      read_binary(&temp, entry_type, bin_mode);
      for(i = 0; i < L2DBL; i++) {
	fp_num.double_b[i] &= temp.double_b[i];
      }
      break;
    case 103:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p -= temp.single_p;
      add_op_list(MINUS, fpbuf, history);
      break;
    case 203:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p -= temp.double_p;
      add_op_list(MINUS, fpbuf, history);
      break;
    case 303:
      read_binary(&temp, entry_type, bin_mode);
      fp_num.single_b |= temp.single_b;
      break;
    case 403:
      read_binary(&temp, entry_type, bin_mode);
      for(i = 0; i < L2DBL; i++) {
	fp_num.double_b[i] |= temp.double_b[i];
      }
      break;
    case 104:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p = temp.single_p - fp_num.single_p;
      add_op_list(MINUSR, fpbuf, history);
      break;
    case 204:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p = temp.double_p - fp_num.double_p;
      add_op_list(MINUSR, fpbuf, history);
      break;
    case 304:
      read_binary(&temp, entry_type, bin_mode);
      fp_num.single_b ^= temp.single_b;
      break;
    case 404:
      read_binary(&temp, entry_type, bin_mode);
      for(i = 0; i < L2DBL; i++) {
	fp_num.double_b[i] |= temp.double_b[i];
      }
      break;
    case 105:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p *= temp.single_p;
      add_op_list(TIMES, fpbuf, history);
      break;
    case 205:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p *= temp.double_p;
      add_op_list(TIMES, fpbuf, history);
      break;
    case 305:
      fp_num.single_b = ~fp_num.single_b;
      break;
    case 405:
      for(i = 0; i < L2DBL; i++) {
	fp_num.double_b[i] = ~fp_num.double_b[i];
      }
      break;
    case 106:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p /= temp.single_p;
      add_op_list(DIV, fpbuf, history);
      break;
    case 206:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p /= temp.double_p;
      add_op_list(DIV, fpbuf, history);
      break;
    case 306:
      fp_num.single_b = fp_num.single_b << 1;
      break;
    case 406:
      overflow = 0x80000000 & fp_num.double_b[0];
      fp_num.double_b[0] = fp_num.double_b[0] << 1;
      for(i = 1; i < L2DBL; i++) {
	long newoverflow = 0x80000000 & fp_num.double_b[i];
	fp_num.double_b[i] = (fp_num.double_b[i] << 1)
	  | ((overflow == 0) ? 0 : 0x80000000);
	overflow = newoverflow;
      }
      break;
    case 107:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p = temp.single_p / fp_num.single_p;
      add_op_list(DIVR, fpbuf, history);
      break;
    case 207:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p = temp.double_p / fp_num.double_p;
      add_op_list(DIVR, fpbuf, history);
      break;
    case 307:
      fp_num.single_b = fp_num.single_b >> 1;
      break;
    case 407:
      overflow = 0x00000001 & fp_num.double_b[L2DBL - 1];
      fp_num.double_b[L2DBL - 1] = fp_num.double_b[L2DBL - 1] >> 1;
      for(i = L2DBL - 2; i >= 0; i++) {
	long newoverflow = 0x00000001 & fp_num.double_b[i];
	fp_num.double_b[i] = (fp_num.double_b[i] >> 1)
	  | ((overflow == 0) ? 0 : 0x00000001);
	overflow = newoverflow;
      }
      break;
    case 108:
      fp_num.single_p = sqrt(fp_num.single_p);
      add_op_list(SQRT, NULL, history);
      break;
    case 208:
      fp_num.double_p = sqrt(fp_num.double_p);
      add_op_list(SQRT, NULL, history);
      break;
    case 109:
      fp_num.single_p = nextafter(fp_num.single_p, fpinf);
      add_op_list(NEXT, NULL, history);
      break;
    case 209:
      fp_num.double_p = nextafter(fp_num.double_p, dpinf);
      add_op_list(NEXT, NULL, history);
      break;
    case 110:
      fp_num.single_p = nextafter(fp_num.single_p, fninf);
      add_op_list(PREV, NULL, history);
      break;
    case 210:
      fp_num.double_p = nextafter(fp_num.double_p, dninf);
      add_op_list(PREV, NULL, history);
      break;
    case 111:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p = (float)pow((double)fp_num.single_p,
				   (double)temp.single_p);
      add_op_list(POW, fpbuf, history);
      break;
    case 211:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p = pow(fp_num.double_p, temp.double_p);
      add_op_list(POW, fpbuf, history);
      break;
    case 112:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p = (float)pow((double)temp.single_p,
				   (double)fp_num.single_p);
      add_op_list(POWR, fpbuf, history);
      break;
    case 212:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p = pow(temp.double_p, fp_num.double_p);
      add_op_list(POWR, fpbuf, history);
      break;
    case 113:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p += (float)pow(2.0, (double)temp.single_p);
      add_op_list(PLUS2, fpbuf, history);
      break;
    case 213:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p += pow(2.0, temp.double_p);
      add_op_list(PLUS2, fpbuf, history);
      break;
    case 114:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p -= (float)pow(2.0, (double)temp.single_p);
      add_op_list(MINUS2, fpbuf, history);
      break;
    case 214:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p -= pow(2.0, temp.double_p);
      add_op_list(MINUS2, fpbuf, history);
      break;
    case 115:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p = (float)pow(2.0, (double)temp.single_p)
	- fp_num.single_p;
      add_op_list(MINUSR2, fpbuf, history);
      break;
    case 215:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p = pow(2.0, temp.double_p) - fp_num.double_p;
      add_op_list(MINUSR2, fpbuf, history);
      break;
    case 116:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p *= (float)pow(2.0, (double)temp.single_p);
      add_op_list(TIMES2, fpbuf, history);
      break;
    case 216:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p *= pow(2.0, temp.double_p);
      add_op_list(TIMES2, fpbuf, history);
      break;
    case 117:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p /= (float)pow(2.0, (double)temp.single_p);
      add_op_list(DIV2, fpbuf, history);
      break;
    case 217:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p /= pow(2.0, temp.double_p);
      add_op_list(DIV2, fpbuf, history);
      break;
    case 118:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.single_p = (float)pow(2.0, (double)temp.single_p)
	/ fp_num.single_p;
      add_op_list(DIVR2, fpbuf, history);
      break;
    case 218:
      read_float(&temp, entry_type, fp_mode, fpbuf);
      fp_num.double_p = pow(2.0, temp.double_p) / fp_num.double_p;
      add_op_list(DIVR2, fpbuf, history);
      break;
    case 50:
      change_round();
      break;
    case 51:
      change_trapping();
      break;
    case 52:
      change_mode(&entry_type);
      break;
    case 53:
      printf("Enter new number of signficant digits:\n");
      scanf("%d", &sigdig);
      break;
    case 54:
      if(entry_type == SINGLEFP || entry_type == DOUBLEFP) {
	fp_mode = (fp_mode == FPSCANF ? FPATOF : FPSCANF);
      }
      else {
	bin_mode = (bin_mode == BIN ? HEX : BIN);
      }
      break;
    case 55:
      if(entry_type == SINGLEFP || entry_type == DOUBLEFP) {
	create_prog(entry_type, history);
      }
      break;
    case 0:
      return;
    default:
      printf("Invalid option selected: %d\n", option);
      option = -1;
    }
    fp_sticky = fpgetsticky();
#ifdef FP_X_INV
    if((fp_sticky & FP_X_INV) != 0) {
      printf("Detected invalid operation exception\n");
    }
#endif
#ifdef FP_X_OFL
    if((fp_sticky & FP_X_OFL) != 0) {
      printf("Detected overflow exception\n");
    }
#endif
#ifdef FP_X_UFL
    if((fp_sticky & FP_X_UFL) != 0) {
      printf("Detected underflow exception\n");
    }
#endif
#ifdef FP_X_DZ
    if((fp_sticky & FP_X_DZ) != 0) {
      printf("Detected divide-by-zero exception\n");
    }
#endif
#ifdef FP_X_IMP
    if((fp_sticky & FP_X_IMP) != 0) {
      printf("Detected loss of precision exception\n");
    }
#endif
#ifdef FP_X_DNML
    if((fp_sticky & FP_X_DNML) != 0) {
      printf("Detected denormalization exception\n");
    }
#endif

    option = -1;
  }
}
    

void print_binary(long number) {
  char string[10];
  char *p;

  sprintf(string, "%08lX", number);
  for(p = string; *p != '\0'; p++) {
    switch(*p) {
    case '0':
      printf("0000");
      break;
    case '1':
      printf("0001");
      break;
    case '2':
      printf("0010");
      break;
    case '3':
      printf("0011");
      break;
    case '4':
      printf("0100");
      break;
    case '5':
      printf("0101");
      break;
    case '6':
      printf("0110");
      break;
    case '7':
      printf("0111");
      break;
    case '8':
      printf("1000");
      break;
    case '9':
      printf("1001");
      break;
    case 'A':
      printf("1010");
      break;
    case 'B':
      printf("1011");
      break;
    case 'C':
      printf("1100");
      break;
    case 'D':
      printf("1101");
      break;
    case 'E':
      printf("1110");
      break;
    case 'F':
      printf("1111");
      break;
    }
  }
}

void fp_handler(int sig, siginfo_t *sip, void *uap) {
  if(sig != SIGFPE) return;
  switch(sip->si_code) {
  case FPE_INTDIV:
    printf("Caught error: Integer divide by zero\n");
    break;
  case FPE_INTOVF:
    printf("Caught error: Integer overflow\n");
    break;
  case FPE_FLTDIV:
    printf("Caught error: Floating point divide by zero\n");
    break;
  case FPE_FLTOVF:
    printf("Caught error: Floating point overflow\n");
    break;
  case FPE_FLTRES:
    printf("Caught error: Floating point inexact result\n");
    break;
  case FPE_FLTINV:
    printf("Caught error: Invalid floating point operation\n");
    break;
  case FPE_FLTSUB:
    printf("Caught error: Subscript out of range\n");
    break;
  default:
    printf("Caught unknown floating point error\n");
  }
}

void read_binary(FP_HACK *num, FP_ENTRY type, BIN_READ mode) {
  char mantbuf[DBL_MANT_DIG + 2];
				/* Used to store the user entered mantissa */
  char expbuf[((sizeof(double) * 8) - DBL_MANT_DIG) + 2];
				/* Used to store the user entered exponent */
  char numbuf[(sizeof(double) * 8) + 2];
				/* Used to store the number in binary */
  char tmpbuf[(sizeof(double) * 8) + 2];
				/* Used as a temporary store when converting
				   from MSB first to LSB first */
  char hexbuf[(sizeof(double) * 2) + 2 + L2DBL];
				/* Used to store the number in hex */
  char signbuf[3];
  int mantsz = (type == SINGLEBIN ? FLT_MANT_DIG : DBL_MANT_DIG) - 1;
				/* Number of binary characters for mantissa */
  int totsz = (type == SINGLEBIN ? sizeof(float) : sizeof(double)) * 8;
				/* Total number of binary characters */
  int expsz = totsz - mantsz - 1;
				/* Number of binary characters for exponent */
  int option = -1;
  char *p, *q, *r, *s;

  if(mode == HEX) {
    if(type == SINGLEBIN) {
      printf("Enter single precision hex number\n");
      fflush(stdin);
      scanf("%lx", &num->single_b);
    }
    else {
      int i;

      printf("Enter double precision hex number "
	     "(separate each word with white space)\n");
      for(i = 0; i < L2DBL; i++) {
	scanf("%lx", &num->double_b[i]);
      }
    }
    return;
  }


  /* Initialise all the character arrays to null chars */
  memset((void *)hexbuf, (int)'\0', sizeof(hexbuf));
  memset((void *)tmpbuf, (int)'\0', sizeof(tmpbuf));
  memset((void *)numbuf, (int)'\0', sizeof(numbuf));
  memset((void *)expbuf, (int)'\0', sizeof(expbuf));
  memset((void *)mantbuf, (int)'\0', sizeof(mantbuf));
  /* Loop to get valid input */
  while(option == -1) {
    int c;

    /* Get the sign */
    printf("Enter sign bit (1 digit)\n");
    fflush(stdin);
    fgets(signbuf, 2, stdin);
    /* Get the exponent */
    printf("Enter exponent bits (%d digits)\n", expsz);
    fflush(stdin);
    fgets(expbuf, expsz + 1, stdin);
    /* Get the mantissa */
    printf("Enter mantissa bits (%d digits)\n", mantsz);
    fflush(stdin);
    fgets(mantbuf, mantsz + 1, stdin);
    /* Find out how the user wants these put into the binary array */
    printf("Choose from the following representation schemes:\n");
    printf("\t[1]\tSign bit, then mantissa, then exponent, MSB then LSB\n");
    printf("\t[2]\tSign bit, then exponent, then mantissa, MSB then LSB\n");
    printf("\t[3]\tSign bit, then mantissa, then exponent, LSB then MSB\n");
    printf("\t[4]\tSign bit, then exponent, then mantissa, LSB then MSB\n");
    printf("\n> ");
    fflush(stdin);
    scanf("%d", &option);
    /* Put them into the binary array as per user selection */
    switch(option) {
    case 1:
      sprintf(numbuf, "%s%s%s", signbuf, mantbuf, expbuf);
      break;
    case 2:
      sprintf(numbuf, "%s%s%s", signbuf, expbuf, mantbuf);
      break;
    case 3:
      sprintf(tmpbuf, "%s%s%s", signbuf, mantbuf, expbuf);
    case 4:
      if(option == 4) sprintf(tmpbuf, "%s%s%s", signbuf, expbuf, mantbuf);
      for(p = tmpbuf, r = numbuf + totsz;
	  p + 8 <= tmpbuf + totsz;
	  p += 8, r -= 8) {
	for(q = p, s = r - 8;
	    q < p + 8; 
	    q++, s++) {
	  (*s) = (*q);
	}
      }
      break;
    default:
      printf("Invalid option selected: %d\n", option);
      option = -1;
    }
    if(option == -1) continue;
    printf("You entered exponent %s (length %d) and mantissa %s (length %d)\n",
	   expbuf, strlen(expbuf), mantbuf, strlen(mantbuf));
    printf("Converted your entry to %s (length %d)\n", numbuf, strlen(numbuf));
    /* Create a hex string containing the binary array in groups of 8 hex
       digits separated by a space */
    for(p = numbuf, s = hexbuf, c = 0;
	p + 4 <= numbuf + totsz; /* Loop through bits 4 at a time */
	p += 4, s++, c++) {
      int i, j = 0;

      /* Separate chucks of longs in hexbuf by spaces */
      if(c % (sizeof(long) * 2) == 0 && c > 0) {
	(*s) = ' ';
	s++;
      }
      /* Loop through this four bits, adding the value of the bit (i) to j */
      for(q = p, i = 8; q < p + 4; i /= 2, q++) {
	if((*q) == '1') j += i;
	else if((*q) != '0') {
	  printf("Invalid character in binary string: [%c]\n", (*q));
	  option = -1;
	  break;
	}
      }
      if(option == -1) break;
      /* Convert j to hex */
      switch(j) {
      case 0: (*s) = '0'; break;
      case 1: (*s) = '1'; break;
      case 2: (*s) = '2'; break;
      case 3: (*s) = '3'; break;
      case 4: (*s) = '4'; break;
      case 5: (*s) = '5'; break;
      case 6: (*s) = '6'; break;
      case 7: (*s) = '7'; break;
      case 8: (*s) = '8'; break;
      case 9: (*s) = '9'; break;
      case 10: (*s) = 'A'; break;
      case 11: (*s) = 'B'; break;
      case 12: (*s) = 'C'; break;
      case 13: (*s) = 'D'; break;
      case 14: (*s) = 'E'; break;
      case 15: (*s) = 'F'; break;
      default:
	fprintf(stderr, "Invalid value for hex digit: %d\n", j);
	abort();
      }
    }
  }
  printf("Hex conversion: %s\n", hexbuf);
  if(type == SINGLEBIN) {
    sscanf(hexbuf, "%lx", &num->single_b);
  }
  else {
    int i;

    for(i = 0; i < L2DBL; i++) {
      sscanf(hexbuf, "%lx", &num->double_b[i]);
    }
  }
}

void read_float(FP_HACK *num, FP_ENTRY type, FP_READ mode, char *entrybuf) {
  fp_except fp_mask = fpgetmask();
  printf("Enter floating point number:\n");
  fpsetmask(0);
  fflush(stdin);
  scanf("%1023s", entrybuf);
  if(type == SINGLEFP) {
    if(mode == FPSCANF) {
      sscanf(entrybuf, "%f", &num->single_p);
      printf("You entered number %*g\n", FLT_DIG, num->single_p);
    }
    else {
      num->single_p = (float)atof(entrybuf);
      printf("You entered number %s -> %*g\n", entrybuf, FLT_DIG,
	     num->single_p);
    }
  }
  else {
    if(mode == FPSCANF) {
      sscanf(entrybuf, "%lf", &num->double_p);
      printf("You entered number %*g\n", DBL_DIG, num->double_p);
    }
    else {
      num->double_p = atof(entrybuf);
      printf("You entered number %s -> %*g\n", entrybuf, DBL_DIG,
	     num->double_p);
    }
  }
  fpsetmask(fp_mask);
}

void change_mode(FP_ENTRY *type) {
  int option = -1;

  while(option == -1) {
    printf("Floating point number hacker -- mode menu\n");
    printf("\n\t[1]\tUse single precision floating point\n");
    printf("\t[2]\tUse double precision floating point\n");
    printf("\t[3]\tUse single precision binary number\n");
    printf("\t[4]\tUse double precision binary number\n");
    printf("\n> ");
    fflush(stdin);
    scanf("%d", &option);
    switch(option) {
    case 1:
      *type = SINGLEFP;
      break;
    case 2:
      *type = DOUBLEFP;
      break;
    case 3:
      *type = SINGLEBIN;
      break;
    case 4:
      *type = DOUBLEBIN;
      break;
    default:
      printf("Invalid entry: %d\n", option);
    }
  }
}

void change_trapping(void) {
  int option = -1;

  while(option == -1) {
    fp_except curx = fpgetmask();

    printf("Floating point number hacker -- trapping control menu\n\n");
#ifdef FP_X_INV
    printf("\t[1]\tTrap invalid operation exception [%s]\n",
	   ((curx & FP_X_INV) != 0) ? "YES" : "NO");
#endif
#ifdef FP_X_OFL
    printf("\t[2]\tTrap overflow exception [%s]\n",
	   ((curx & FP_X_OFL) != 0) ? "YES" : "NO");
#endif
#ifdef FP_X_UFL
    printf("\t[3]\tTrap underflow exception [%s]\n",
	   ((curx & FP_X_UFL) != 0) ? "YES" : "NO");
#endif
#ifdef FP_X_DZ
    printf("\t[4]\tTrap divide-by-zero exception [%s]\n",
	   ((curx & FP_X_DZ) != 0) ? "YES" : "NO");
#endif
#ifdef FP_X_IMP
    printf("\t[5]\tTrap loss of precision exception [%s]\n",
	   ((curx & FP_X_IMP) != 0) ? "YES" : "NO");
#endif
#ifdef FP_X_DNML
    printf("\t[6]\tTrap denormalization exception\n",
	   ((curx & FP_X_DNML) != 0) ? "YES" : "NO");
#endif
    printf("\n\t[0]\tReturn to entry menu\n");
    printf("\n> ");
    fflush(stdin);
    scanf("%d", &option);
    switch(option) {
#ifdef FP_X_INV
    case 1:
      curx ^= FP_X_INV;
      break;
#endif
#ifdef FP_X_OFL
    case 2:
      curx ^= FP_X_OFL;
      break;
#endif
#ifdef FP_X_UFL
    case 3:
      curx ^= FP_X_UFL;
      break;
#endif
#ifdef FP_X_DZ
    case 4:
      curx ^= FP_X_DZ;
      break;
#endif
#ifdef FP_X_IMP
    case 5:
      curx ^= FP_X_IMP;
      break;
#endif
#ifdef FP_X_DNML
    case 6:
      curx ^= FP_X_DNML;
      break;
#endif
    case 0:
      return;
    default:
      printf("Invalid entry: %d -- try again\n", option);
    }
    fpsetmask(curx);
    option = -1;
  }
}

void change_round(void) {
  int option = -1;
  
  while(option == -1) {
    fp_rnd curmode = fpgetround();
    printf("Floating point number hacker -- rounding control menu\n\n");
    printf("\t[1]\tRound to nearest representable number (tie -> even)\n");
    printf("\t[2]\tRound to plus infinity\n");
    printf("\t[3]\tRound to minus infinity\n");
    printf("\t[4]\tRound to zero (truncate)\n");
    printf("\nCurrent setting: ");
    switch(curmode) {
    case FP_RN:
      printf("[1]\n");
      break;
    case FP_RP:
      printf("[2]\n");
      break;
    case FP_RM:
      printf("[3]\n");
      break;
    case FP_RZ:
      printf("[4]\n");
      break;
    default:
      printf("Unrecognised mode %d\n", (int)curmode);
    }
    printf("\n> ");
    fflush(stdin);
    scanf("%d", &option);
    switch(option) {
    case 1:
      fpsetround(FP_RN);
      break;
    case 2:
      fpsetround(FP_RP);
      break;
    case 3:
      fpsetround(FP_RM);
      break;
    case 4:
      fpsetround(FP_RZ);
      break;
    default:
      printf("Invalid entry: %d\n", option);
      option = -1;
    }
  }
}

void create_prog(FP_ENTRY type, OP_LIST *commands) {
  char tmpc[1024], tmpx[1024], cmd[4096], *gccout;
  FILE *fp;
  long i;
  char *precision = (type == SINGLEFP ? "float" : "double");
  char *precstr = (type == SINGLEFP ? "e" : "le");
  struct stat fstat;
  int useSignal;

  printf("Do you want to use signals [enter y]? (alternative: sticky bits)\n");
  fflush(stdin);
  useSignal = fgetc(stdin);
  useSignal = tolower(useSignal);
  if(useSignal == 'y') {
    printf("Using signals\n");
  }
  else {
    printf("Using sticky bit testing\n");
  }
  i = 0;
  do {
    i++;
    sprintf(tmpc, "%s/FPHK%08ld.c", P_tmpdir, i);
  } while(stat(tmpc, &fstat) == 0);
  sprintf(tmpx, "%s/FPHK%08ld", P_tmpdir, i);
  fp = fopen(tmpc, "w");
  if(fp == NULL) {
    fprintf(stderr, "Error creating temporary C file:\n");
    perror(tmpc);
    return;
  }
  fprintf(fp, "static const %s pinf = 1.0%s / 0.0%s;\n", precision,
          (type == SINGLEFP ? "F" : ""));
  fprintf(fp, "static const %s ninf = -1.0%s / 0.0%s;\n", precision,
          (type == SINGLEFP ? "F" : ""));
  if(useSignal == 'y') {
    fprintf(fp,
	    "#include <signal.h>\n"
	    "#include <ieeefp.h>\n"
	    "#include <errno.h>\n"
	    "#include <math.h>\n"
	    "void fph(int sig, siginfo_t *sip, void *uap) {\n"
	    "  if(sig != SIGFPE) return;\n"
	    "  switch(sip->si_code) {\n"
	    "  case FPE_INTDIV: printf(\"Caught int div zero\\n\"); break;\n"
	    "  case FPE_INTOVF: printf(\"Caught int overflow\\n\"); break;\n"
	    "  case FPE_FLTDIV: printf(\"Caught flt div zero\\n\"); break;\n"
	    "  case FPE_FLTOVF: printf(\"Caught flt overflow\\n\"); break;\n"
	    "  case FPE_FLTRES: printf(\"Caught flt inexact\\n\"); break;\n"
	    "  case FPE_FLTINV: printf(\"Caught flt invalid op\\n\"); break;\n"
	    "  case FPE_FLTSUB: printf(\"Caught subscrpt out rng\\n\");\n"
	    "    break;\n"
	    "  default: printf(\"Caught unknown FPE\\n\");\n"
	    "  }\n"
	    "}\n"
	    "int main(int argc, char **argv) {\n"
	    "  %s var;\n"
	    "  struct sigaction fpa;\n"
	    "  fpsetmask(\n"
	    "#ifdef FP_X_INV\n"
	    "            FP_X_INV |\n"
	    "#endif\n"
	    "#ifdef FP_X_DNML\n"
	    "            FP_X_DNML |\n"
	    "#endif\n"
	    "#ifdef FP_X_DZ\n"
	    "            FP_X_DZ |\n"
	    "#endif\n"
	    "#ifdef FP_X_OFL\n"
	    "            FP_X_OFL |\n"
	    "#endif\n"
	    "#ifdef FP_X_UFL\n"
	    "            FP_X_UFL |\n"
	    "#endif\n"
	    "#ifdef FP_X_IMP\n"
	    "            FP_X_IMP |\n"
	    "#endif\n"
	    "            0);\n"
	    "  fpa.sa_flags = SA_SIGINFO;\n"
	    "  fpa.sa_sigaction = fph;\n"
	    "  if(sigemptyset(&fpa.sa_mask) == -1) {\n"
	    "    perror(\"Setting up floating point error signal mask\");\n"
	    "    abort();\n"
	    "  }\n"
	    "  if(sigaction(SIGFPE, &fpa, NULL) == -1) {\n"
	    "    perror(\"Setting up floating-point error signal \""
	    "           \"trapping\");\n"
	    "    abort();\n"
	    "  }\n"
	    , precision);
  }
  else {
    fprintf(fp, 
	    "#include <math.h>\n"
	    "#include <ieeefp.h>\n"
	    "void fps(void) {\n"
	    "  fp_except fp_sticky;\n"
	    "  fp_sticky = fpgetsticky();\n"
	    "#ifdef FP_X_INV\n"
	    "  if((fp_sticky & FP_X_INV) != 0) {\n"
	    "    printf(\"Detected invalid operation exception\\n\");\n"
	    "  }\n"
	    "#endif\n"
	    "#ifdef FP_X_OFL\n"
	    "  if((fp_sticky & FP_X_OFL) != 0) {\n"
	    "    printf(\"Detected overflow exception\\n\");\n"
	    "  }\n"
	    "#endif\n"
	    "#ifdef FP_X_UFL\n"
	    "  if((fp_sticky & FP_X_UFL) != 0) {\n"
	    "    printf(\"Detected underflow exception\\n\");\n"
	    "  }\n"
	    "#endif\n"
	    "#ifdef FP_X_DZ\n"
	    "  if((fp_sticky & FP_X_DZ) != 0) {\n"
	    "    printf(\"Detected divide-by-zero exception\\n\");\n"
	    "  }\n"
	    "#endif\n"
	    "#ifdef FP_X_IMP\n"
	    "  if((fp_sticky & FP_X_IMP) != 0) {\n"
	    "    printf(\"Detected loss of precision exception\\n\");\n"
	    "  }\n"
	    "#endif\n"
	    "#ifdef FP_X_DNML\n"
	    "  if((fp_sticky & FP_X_DNML) != 0) {\n"
	    "    printf(\"Detected denormalization exception\\n\");\n"
	    "  }\n"
	    "#endif\n"
	    "  fpsetsticky(0);\n"
	    "}\n"
	    "int main(int argc, char **argv) {\n"
	    "  %s var;\n"
	    "  fpsetsticky(0);\n"
	    , precision);
  }
  while(commands != NULL) {
    switch(commands->op) {
    case SET:
      fprintf(fp,
	      "  printf(\"Setting var to %s...\\n\");\n"
	      "  var = %s;\n"
	      , commands->operand, commands->operand);
      break;
    case PLUS:
      fprintf(fp,
	      "  printf(\"Adding %s to var...\\n\");\n"
	      "  var += %s;\n"
	      , commands->operand, commands->operand);
      break;
    case MINUS:
      fprintf(fp,
	      "  printf(\"Subtracting %s from var...\\n\");\n"
	      "  var -= %s;\n"
	      , commands->operand, commands->operand);
      break;
    case MINUSR:
      fprintf(fp,
	      "  printf(\"Subtracting var from %s...\\n\");\n"
	      "  var = %s - var;\n"
	      , commands->operand, commands->operand);
      break;
    case TIMES:
      fprintf(fp,
	      "  printf(\"Multiplying var by %s...\\n\");\n"
	      "  var *= %s;\n"
	      , commands->operand, commands->operand);
      break;
    case DIV:
      fprintf(fp,
	      "  printf(\"Dividing var by %s...\\n\");\n"
	      "  var /= %s;\n"
	      , commands->operand, commands->operand);
      break;
    case DIVR:
      fprintf(fp,
	      "  printf(\"Dividing %s by var...\\n\");\n"
	      "  var = %s / var;\n"
	      , commands->operand, commands->operand);
      break;
    case SQRT:
      fprintf(fp,
	      "  printf(\"Taking the square root of var...\\n\");\n"
	      "  var = sqrt(var);\n"
	      );
      break;
    case NEXT:
      fprintf(fp,
              "  printf(\"Getting the next after var...\\n\");\n"
              "  var = nextafter(var, pinf);\n"
              );
      break;
    case PREV:
      fprintf(fp,
              "  printf(\"Getting the previous before var...\\n\");\n"
              "  var = nextafter(var, ninf);\n"
              );
      break;
    case POW:
      fprintf(fp,
	      "  printf(\"Raising var to the power %s...\\n\");\n"
	      "  var = pow(var, %s);\n"
	      , commands->operand, commands->operand);
      break;
    case POWR:
      fprintf(fp,
	      "  printf(\"Raising %s to the power var...\\n\");\n"
	      "  var = pow(%s, var);\n"
	      , commands->operand, commands->operand);
      break;
    case PLUS2:
      fprintf(fp,
	      "  printf(\"Adding 2^%s to var...\\n\");\n"
	      "  var += pow(2.0, %s);\n"
	      , commands->operand, commands->operand);
      break;
    case MINUS2:
      fprintf(fp,
	      "  printf(\"Subtracting 2^%s from var...\\n\");\n"
	      "  var -= pow(2.0, %s);\n"
	      , commands->operand, commands->operand);
      break;
    case MINUSR2:
      fprintf(fp,
	      "  printf(\"Subtracting var from 2^%s...\\n\");\n"
	      "  var = pow(2.0, %s) - var;\n"
	      , commands->operand, commands->operand);
      break;
    case TIMES2:
      fprintf(fp,
	      "  printf(\"Multiplying var by 2^%s...\\n\");\n"
	      "  var *= pow(2.0, %s);\n"
	      , commands->operand, commands->operand);
      break;
    case DIV2:
      fprintf(fp,
	      "  printf(\"Dividing var by 2^%s...\\n\");\n"
	      "  var /= pow(2.0, %s);\n"
	      , commands->operand, commands->operand);
      break;
    case DIVR2:
      fprintf(fp,
	      "  printf(\"Dividing 2^%s by var...\\n\");\n"
	      "  var = pow(2.0, %s) / var;\n"
	      , commands->operand, commands->operand);
      break;
    default:
      fprintf(stderr, "Invalid operator in history: %d\n", commands->op);
      commands = commands->next;
      continue;
    }
    if(useSignal != 'y') {
      fprintf(fp,
	      "  fps();\n");
    }
    fprintf(fp,
	    "  printf(\"...done. var is now:\\n%%.%d%s\\n\", var);\n"
	    , PROGACC, precstr);
    commands = commands->next;
  }
  fprintf(fp,
	  "  return 0;\n"
	  "}\n");
  fclose(fp);
  gccout = tempnam(NULL, "FPHKO");
  if(gccout == NULL) {
    perror("Failed to obtain temporary output file name");
    if(unlink(tmpc) != 0) {
      fprintf(stderr, "Failed to delete temporary C file ");
      perror(tmpc);
    }
    return;
  }
  sprintf(cmd, "%s -o %s %s -lm > %s 2>&1", COMPILER, tmpx, tmpc, gccout);
  if(system(cmd) < 0) {
    fprintf(stderr, "Compiler command failed to execute.\n");
    perror(cmd);
  }
  else if(stat(tmpx, &fstat) == 0) {
    system(tmpx);
    if(unlink(tmpx) != 0) {
      fprintf(stderr, "Failed to delete temporary executable ");
      perror(tmpx);
    }
    if(unlink(gccout) != 0) {
      fprintf(stderr, "Failed to delete compiler command output file ");
      perror(gccout);
    }
  }
  else {
    fprintf(stderr, "Compiler command failed. Command and output follow:\n"
	    "%s\n", cmd);
    fp = fopen(gccout, "r");
    if(fp == NULL) {
      fprintf(stderr, "Output unavailable. ");
      perror(gccout);
    }
    else {
      while(!feof(fp)) {
	fputc(fgetc(fp), stderr);
      }
      fprintf(stderr, "\n");
      fclose(fp);
    }
    if(unlink(gccout) != 0) {
      fprintf(stderr, "Failed to delete compiler command output file ");
      perror(gccout);
    }
  }
  if(unlink(tmpc) != 0) {
    fprintf(stderr, "Failed to delete temporary C file ");
    perror(tmpc);
  }
  free(gccout);
}
